/*
 * Copyright (c) 2025      Andreas Schneider <asn@samba.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

use clap::Parser;
use clap::Subcommand;
use std::process::ExitCode;
use tracing::Level;
use tracing_subscriber::FmtSubscriber;

#[derive(Parser)]
#[command(name = "kirmesctl", version, about, long_about = None)]
struct Cli {
    /// Service to use
    #[arg(long, default_value_t = String::from("io.systemd.Multiplexer"))]
    service: String,
    /// Tracing level
    #[arg(long, value_parser = parse_tracing_level)]
    tracing: Option<Level>,

    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    /// Inspect user(s)
    User { user: Option<String> },
    /// Inspect group(s)
    Group { group: Option<String> },
    /// Show users that are members of specified groups
    UsersInGroup { group: Option<String> },
    /// Show users that are members of specified groups
    GroupsOfUser { user: Option<String> },
}

fn parse_tracing_level(lvl: &str) -> Result<tracing::Level, String> {
    match lvl.to_lowercase().as_str() {
        "warn" => return Ok::<Level, String>(Level::WARN),
        "info" => return Ok(Level::INFO),
        "debug" => return Ok(Level::DEBUG),
        "trace" => return Ok(Level::TRACE),
        &_ => return Ok(Level::ERROR),
    };

    #[allow(unreachable_code)]
    Err(String::from(
        "Tracing level must be one of the following values: error|warn|info|debug|trace",
    ))
}

fn main() -> ExitCode {
    let cli = Cli::parse();

    if let Some(lvl) = cli.tracing {
        let subscriber = FmtSubscriber::builder().with_max_level(lvl).finish();
        if tracing::subscriber::set_global_default(subscriber).is_err() {
            eprintln!("Failed to set tracing level");
            return ExitCode::FAILURE;
        }
    }

    match &cli.command {
        Commands::User { user } => {
            let mut user = user.clone();
            let mut uid = None;

            if let Some(u) = &user {
                uid = match u.parse::<u32>() {
                    Ok(ok) => {
                        user = None;
                        Some(ok)
                    },
                    Err(_) => None,
                };
            };

            let user_records = match kirmes::get_user_record(uid, user, cli.service) {
                Ok(user_records) => user_records,
                Err(e) => {
                    println!("Failed to lookup user record: {e:?}");
                    return ExitCode::FAILURE;
                },
            };

            for record in user_records {
                println!("{record}");
            }
        },
        Commands::Group { group } => {
            let mut group = group.clone();
            let mut gid = None;

            if let Some(g) = &group {
                gid = match g.parse::<u32>() {
                    Ok(ok) => {
                        group = None;
                        Some(ok)
                    },
                    Err(_) => None,
                };
            };

            let group_records = match kirmes::get_group_record(gid, group, cli.service) {
                Ok(group_records) => group_records,
                Err(e) => {
                    println!("Failed to lookup group record: {e:?}");
                    return ExitCode::FAILURE;
                },
            };

            for record in group_records {
                println!("{record}");
            }
        },
        Commands::UsersInGroup { group } => {
            let usr = None;
            let grp = group.clone();

            let memberships = match kirmes::get_memberships(usr, grp, cli.service) {
                Ok(group_records) => group_records,
                Err(e) => {
                    if let Some(g) = group {
                        println!("Failed to lookup memberships of group '{g}': {e:?}");
                    } else {
                        println!("Failed to lookup memberships: {e:?}");
                    }
                    return ExitCode::FAILURE;
                },
            };

            println!("USER\tGROUP");
            for m in &memberships {
                println!("{}\t{}", m.user_name, m.group_name);
            }
            println!("\n{} memberships listed", memberships.len());
        },
        Commands::GroupsOfUser { user } => {
            let usr = user.clone();
            let grp = None;

            let memberships = match kirmes::get_memberships(usr, grp, cli.service) {
                Ok(group_records) => group_records,
                Err(e) => {
                    if let Some(u) = user {
                        println!("Failed to lookup memberships of user '{u}': {e:?}");
                    } else {
                        println!("Failed to lookup memberships: {e:?}");
                    }
                    return ExitCode::FAILURE;
                },
            };

            println!("USER\tGROUP");
            for m in &memberships {
                println!("{}\t{}", m.user_name, m.group_name);
            }
            println!("\n{} memberships listed", memberships.len());
        },
    }

    ExitCode::SUCCESS
}
