/*
 * Copyright (c) 2025      Andreas Schneider <asn@samba.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

//! # C API for kirmes
//!
//! [kirmes](https://gitlab.com/kirmes/kirmes/) is a library written in Rust for accessing
//! systemd's userdb.
//!
//! This is the C-compatible API

// Basically everything will be unsafe since this is a FFI
#![allow(clippy::undocumented_unsafe_blocks)]

use crate::Error;
pub use crate::c_user_record::*;
use crate::get_user_record;
use std::ffi::CStr;
use std::os::raw::{c_char, c_uint};

/// This type represents success and errors working with systemd's userdb.
#[repr(C)]
#[allow(clippy::enum_variant_names)]
pub enum KirmesStatus {
    /// Normal operation
    KirmesStatusSuccess = 0,

    /// Incorrectly set service field
    KirmesStatusBadService,
    /// In case uid/gid and name are specified and two records have been found
    KirmesStatusConflictingRecordFound,
    /// If the service doesn't support enumeration
    KirmesStatusEnumerationNotSupported,
    /// An internal error occurred
    KirmesStatusInternalError,
    /// An invalid parameter has been passed
    KirmesStatusInvalidParameter,
    /// An invalid record has been found, e.g. json parsing failed.
    KirmesStatusInvalidRecord,
    /// If no user record matching the specified UID or name is known
    KirmesStatusNoRecordFound,
    /// The backing subsystem is not operational
    KirmesStatusServiceNotAvailable,
}

impl From<Error> for KirmesStatus {
    fn from(err: Error) -> Self {
        match err {
            Error::BadService => KirmesStatus::KirmesStatusBadService,
            Error::ConflictingRecordFound => KirmesStatus::KirmesStatusConflictingRecordFound,
            Error::EnumerationNotSupported => KirmesStatus::KirmesStatusEnumerationNotSupported,
            Error::InternalError => KirmesStatus::KirmesStatusInternalError,
            Error::InvalidRecord => KirmesStatus::KirmesStatusInvalidRecord,
            Error::NoRecordFound => KirmesStatus::KirmesStatusNoRecordFound,
            Error::ServiceNotAvailable => KirmesStatus::KirmesStatusServiceNotAvailable,
        }
    }
}

fn char_to_option_string(s: *const c_char) -> Option<String> {
    if s.is_null() {
        return None;
    }

    let c_str = unsafe { CStr::from_ptr(s) };
    match c_str.to_str() {
        Ok(slice) => Some(slice.to_string()),
        Err(_) => None,
    }
}

fn uint_ptr_to_option_u32(i: *const c_uint) -> Option<u32> {
    if i.is_null() {
        return None;
    }

    unsafe { Some(*i) }
}

/// The method looks up or enumerates a user record.
///
/// * `uid`: Specifies the numeric UNIX UID to search for.
/// * `user_name`: Specifies the name of the user to search for.
/// * `service`: Mandatory argument that should be set to the service name being talked to.
///              If you don't want to talk to a specific service, use: `io.systemd.Multiplexer`
/// * `out`: An array of `struct UserRecord`
///
/// # Returns
/// On Success: ``KirmesStatusSuccess``
/// On Error: ``KirmesStatusBadService``
///           ``KirmesStatusConflictingRecordFound``
///           ``KirmesStatusEnumerationNotSupported``
///           ``KirmesStatusInternalError``
///           ``KirmesStatusInvalidParameter``
///           ``KirmesStatusInvalidRecord``
///           ``KirmesStatusNoRecordFound``
///           ``KirmesStatusServiceNotAvailable``
///
#[no_mangle]
pub unsafe extern "C" fn kirmes_get_user_record(
    uid: *const c_uint,
    user_name: *const c_char,
    service: *const c_char,
    out_records: *mut *mut KirmesUserRecord,
    num_records: *mut usize,
) -> KirmesStatus {
    let r_uid = uint_ptr_to_option_u32(uid);
    let r_user_name = char_to_option_string(user_name);
    let Some(r_service) = char_to_option_string(service) else {
        return KirmesStatus::KirmesStatusInvalidParameter;
    };

    let user_records = match get_user_record(r_uid, r_user_name, r_service) {
        Ok(r) => r,
        Err(e) => return e.into(),
    };

    let mut out: Vec<KirmesUserRecord> = user_records
        .into_iter()
        .map(std::convert::Into::into)
        .collect();

    out.shrink_to_fit(); // Don't waste space
    if !num_records.is_null() {
        *num_records = out.len();
    }

    if !out_records.is_null() {
        out_records.write(Box::into_raw(out.into_boxed_slice()).cast());
    }

    KirmesStatus::KirmesStatusSuccess
}

#[no_mangle]
/// Free a `KirmesUserRecord` array.
///
/// * `records`: The pointer to free.
pub extern "C" fn kirmes_user_record_free(records: *mut KirmesUserRecord) {
    if !records.is_null() {
        unsafe {
            let _ = Box::from_raw(records);
        }
    }
}
