/*
 * Copyright (c) 2025      Andreas Schneider <asn@samba.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

use derivative::Derivative;
use serde::Deserialize;
use serde::Deserializer;
use serde_json::Value;
use std::collections::HashMap;
use tracing::error;

/// Disposition of the user, i.e. the context it is defined in.
#[derive(Derivative, Deserialize)]
#[derivative(Default(bound = ""), Debug)]
pub enum UserDisposition {
    /// root and nobody
    #[serde(rename = "intrinsic")]
    UserIntrinsic,
    /// statically allocated users for system services
    #[serde(rename = "system")]
    UserSystem,
    /// dynamically allocated users for system services
    #[serde(rename = "dynamic")]
    UserDynamic,
    /// regular (typically human users)
    #[derivative(Default)]
    #[serde(rename = "regular")]
    UserRegular,
    /// UID ranges allocated for container uses
    #[serde(rename = "container")]
    UserContainer,
    /// UID range allocated for foreign OS images
    #[serde(rename = "foreign")]
    UserForeign,
    /// Range above 2^31
    #[serde(rename = "reserved")]
    UserReserved,
    _UserDispositionMax,
}

/// Indicates the storage mechanism for the user’s home directory.
#[derive(Derivative, Deserialize)]
#[derivative(Default(bound = ""), Debug)]
pub enum UserHomeStorage {
    #[serde(rename = "classic")]
    #[derivative(Default)]
    UserClassic,
    #[serde(rename = "luks")]
    UserLuks,
    #[serde(rename = "directory")]
    UserDirectory,
    #[serde(rename = "subvolume")]
    UserSubvolume,
    #[serde(rename = "fscrypt")]
    UserFscrypt,
    #[serde(rename = "cifs")]
    UserCifs,
    _UserStorageMax,
}

/// Controls whether the home area shall be grown automatically
#[derive(Derivative, Deserialize)]
#[derivative(Default(bound = ""), Debug)]
pub enum AutoResizeMode {
    #[derivative(Default)]
    #[serde(rename = "off")]
    AutoResizeOff,
    #[serde(rename = "grow")]
    AutoResizeGrow,
    #[serde(rename = "shrink-and-grow")]
    AutoResizeShrinkAndGrow,
    _AutoResizeModeMax,
}

/// The `EncryptedKey` for PKCS#11
#[derive(Derivative, Deserialize)]
#[derivative(Default(bound = ""), Debug)]
pub struct Pkcs11EncryptedKey {
    /// PKCS#11 security token URI,
    pub uri: String,
    /// UNIX password hash to test the key against.
    #[serde(rename = "hashedPassword")]
    pub hashed_password: String,
    /// Base64-encoded encrypted key
    pub data: String,
}

/// Authentication support with FIDO2 devices that implement the hmac-secret extension
#[derive(Derivative, Deserialize)]
#[derivative(Default(bound = ""), Debug)]
pub struct Fido2HmacSalt {
    pub credential: String,
    pub salt: String,
    #[serde(rename = "hashedPassword")]
    pub hashed_password: String,
    pub uv: bool,
    pub up: bool,
    #[serde(rename = "clientPin")]
    pub client_pin: bool,
}

/// Recovery keys for the user account
#[derive(Derivative, Deserialize)]
#[derivative(Default(bound = ""), Debug)]
pub struct RecoveryKey {
    /// The type of recovery key, must be "modhex64" right now
    #[serde(rename = "type")]
    pub key_type: String,
    /// A UNIX password hash of the normalized form of modhex64
    #[serde(rename = "hashedPassword")]
    pub hashed_password: String,
}

/// Contains information that the user and the administrator should be able to see, but other users
/// should not.
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct UserPrivileged {
    /// A user-selected password hint in free-form text.
    #[serde(default, rename = "passwordHint")]
    pub password_hint: Option<String>,
    /// An array of strings, each containing a hashed UNIX password string, in the format crypt(3)
    /// generates.
    #[serde(default, rename = "hashedPassword")]
    pub hashed_passwords: Option<Vec<String>>,
    /// An array of strings, each listing an SSH public key that is authorized to access the
    /// account.
    #[serde(default, rename = "sshAuthorizedKeys")]
    pub ssh_authorized_keys: Option<Vec<String>>,
    /// An array of `Pkcs11EncryptedKey` objects.
    #[serde(default, rename = "Pkcs11EncryptedKey")]
    pub pkcs11_encrypted_key: Option<Vec<Pkcs11EncryptedKey>>,
    /// An array of objects, implementing authentication support with FIDO2 devices that implement
    /// the hmac-secret extension.
    #[serde(default, rename = "Fido2HmacSalt")]
    pub fido2_hmac_salt: Option<Vec<Fido2HmacSalt>>,
    /// An array of objects, each defining a recovery key.
    #[serde(default, rename = "recoveryKey")]
    pub recovery_key: Option<Vec<RecoveryKey>>,
}

/// Resource management-related fields
#[allow(clippy::struct_excessive_bools)]
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct UserPerMachine {
    /// An array of strings that are formatted 128-bit IDs in hex.
    #[serde(default)]
    pub match_machine_id: Option<Vec<String>>,
    /// An array of strings that are valid hostnames.
    #[serde(default)]
    pub match_hostname: Option<Vec<String>>,

    /// The absolute path to a world-readable copy of the user’s blob directory.
    #[serde(default, rename = "blobDirectory")]
    pub blob_directory: Option<String>,
    /// An object, which maps valid blob directory filenames to SHA256 hashes formatted as hex
    /// strings.
    #[serde(
        default,
        rename = "blobManifest",
        deserialize_with = "decode_string_hashmap"
    )]
    pub blob_manifest: Option<HashMap<String, String>>,
    /// The name of an icon picked by the user.
    #[serde(default, rename = "iconName")]
    pub icon_name: Option<String>,
    /// A free-form location string describing the location of the user.
    #[serde(default)]
    pub location: Option<String>,
    /// The shell binary to use for terminal logins of this user.
    #[serde(default)]
    pub shell: Option<String>,
    /// The `umask` to set for the user’s login sessions.
    #[serde(default)]
    pub umask: Option<u32>,
    /// An array of strings, each containing an environment variable and its value to set for the
    /// user’s login session.
    #[serde(default, deserialize_with = "decode_string_hashmap")]
    pub environment: Option<HashMap<String, String>>,
    /// The preferred timezone of the user.
    #[serde(default, rename = "timeZone")]
    pub time_zone: Option<String>,
    /// The preferred language/locale for the user.
    #[serde(default, rename = "preferredLanguage")]
    pub preferred_language: Option<String>,
    /// The preferred languages/locales that should be used in the event that translations for the
    /// `preferredLanguage` are missing
    #[serde(default, rename = "additionalLanguages")]
    pub additional_languages: Option<Vec<String>>,
    /// An integer value in the range -20…19
    #[serde(default, rename = "niceLevel")]
    pub nice_level: Option<i32>,
    // An object, where each key refers to a Linux resource limit (such as RLIMIT_NOFILE and
    // similar).
    // #[serde(default, rename = "resourceLimits")]
    // TODO pub resource_limits: ???,
    /// If true, the user account is locked, the user may not log in.
    #[derivative(Default(value = "false"))]
    #[serde(default)]
    pub locked: bool,
    /// A timestamp when record should be considered invalid for the purpose of logging in (µs
    /// since the UNIX epoch 1970).
    #[serde(default, rename = "notBeforeUSec")]
    pub not_before_usec: Option<u64>,
    /// The point in time after which logins shall not be permitted anymore (µs since the
    /// UNIX epoch 1970).
    #[serde(default, rename = "notAfterUSec")]
    pub not_after_usec: Option<u64>,
    /// The storage mechanism for the user’s home directory.
    #[derivative(Default(value = "UserHomeStorage::UserClassic"))]
    #[serde(default)]
    pub storage: UserHomeStorage,
    /// The intended home directory disk space in bytes to assign to the user.
    #[serde(default, rename = "diskSize")]
    pub disk_size: Option<u64>,
    /// Similar to diskSize but takes a relative value, but specifies a fraction of the available
    /// disk space on the selected storage medium to assign to the user.
    #[serde(default, rename = "diskSizeRelative")]
    pub disk_size_relative: Option<u64>,
    /// The absolute path to the skeleton directory to populate a new home directory from.
    #[serde(default, rename = "skeletonDirectory")]
    pub skeleton_directory: Option<String>,
    /// The UNIX access mask in the range 0…511 for the home directory when it is first created.
    #[serde(default, rename = "accessMask")]
    pub access_mask: Option<u32>,
    /// The maximum number of tasks the user may start in parallel during system runtime.
    #[serde(default, rename = "taskMax")]
    pub tasks_max: Option<u64>,
    /// Upper memory limits for all processes of the user.
    #[serde(default, rename = "memoryHigh")]
    pub memory_high: Option<u64>,
    /// Upper memory limits for all processes of the user.
    #[serde(default, rename = "memoryMax")]
    pub memory_max: Option<u64>,
    /// CPU scheduling weights for the user’s processes (range 1…10000 ,defaults to
    /// 100).
    #[serde(default, rename = "cpuWeight")]
    pub cpu_weight: Option<u32>,
    /// IO scheduling weights for the user’s processes (range 1…10000 ,defaults to 100).
    #[serde(default, rename = "ioWeight")]
    pub io_weight: Option<u32>,
    /// Control nosuid mount flag of the user’s home directories.
    #[serde(default, rename = "mountNoSuid")]
    pub mount_no_suid: bool,
    /// Control nodev mount flag of the user’s home directories.
    #[serde(default, rename = "mountNoDev")]
    pub mount_no_dev: bool,
    /// Control noexec mount flag of the user’s home directories.
    #[serde(default, rename = "mountNoExec")]
    pub mount_no_exec: bool,
    /// The Windows File Sharing domain (CIFS) to use.
    #[serde(default, rename = "cifsDomain")]
    pub cifs_domain: Option<String>,
    /// The Windows File Sharing user name (CIFS) to associate this user record with.
    #[serde(default, rename = "cifsName")]
    pub cifs_user_name: Option<String>,
    /// The Windows File Share service (CIFS) to mount as home directory of the user on login.
    #[serde(default, rename = "cifsService")]
    pub cifs_service: Option<String>,
    /// Additional mount options to pass to mount.cifs when mounting the home directory CIFS share.
    #[serde(default, rename = "cifsExtraMountOptions")]
    pub cifs_extra_mount_options: Option<String>,
    /// An absolute file system path to the file, directory or block device to use for storage
    /// backing the home directory.
    #[serde(default, rename = "imagePath")]
    pub image_path: Option<String>,
    /// The numeric UNIX user ID (UID) to use for the user.
    #[serde(default)]
    pub uid: Option<u32>,
    /// The numeric UNIX group ID (GID) to use for the user.
    #[serde(default)]
    pub gid: Option<u32>,
    /// An array of strings, each indicating a UNIX group this user shall be a member of.
    #[serde(default, rename = "memberOf")]
    pub member_of: Option<Vec<String>>,
    /// One of ext4, xfs, btrfs (possibly others) to use as file system for the user’s home
    /// directory.
    #[serde(default, rename = "fileSystemType")]
    pub file_system_type: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the GPT partition UUID the home directory is
    /// located in.
    #[serde(default, rename = "partitionUuid")]
    pub partition_uuid: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the LUKS volume UUID the home directory is
    /// located in.
    #[serde(default, rename = "luksUuid")]
    pub luks_uuid: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the file system UUID the home directory is
    /// located in.
    #[serde(default, rename = "fileSystemUuid")]
    pub file_system_uuid: Option<String>,
    ///  If true and luks storage is used, controls whether the loopback block devices, LUKS and
    ///  the file system on top shall be used in discard mode.
    #[serde(default, rename = "luksDiscard")]
    pub luks_discard: bool,
    /// Similar to luksDiscard, it controls whether to trim/allocate the file system/backing file
    /// when deactivating the home directory.
    #[serde(default, rename = "luksOfflineDiscard")]
    pub luks_offline_discard: bool,
    /// The cipher to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksCipher")]
    pub luks_cipher: Option<String>,
    // Selects the cipher mode to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksCipherMode")]
    pub luks_cipher_mode: Option<String>,
    /// The volume key length in bytes to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksVolumeKeySize")]
    pub luks_volume_key_size: Option<u64>,
    /// Selects the hash algorithm to use for the PBKDF operation.
    #[serde(default, rename = "luksPbkdfHashAlgorithm")]
    pub luks_pbkdf_hash_algorithm: Option<String>,
    /// The PBKDF type to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksPbkdfType")]
    pub luks_pbkdf_type: Option<String>,
    /// The intended number of iterations for the PBKDF operation.
    #[serde(default, rename = "luksPbkdfForceIterations")]
    pub luks_pbkdf_force_iterations: Option<u64>,
    /// The intended time cost for the PBKDF operation, in µs.
    #[serde(default, rename = "luksPbkdfTimeCostUSec")]
    pub luks_pbkdf_time_cost_usec: Option<u64>,
    /// The intended memory cost for the PBKDF operation, in bytes.
    #[serde(default, rename = "luksPbkdfMemoryCost")]
    pub luks_pbkdf_memory_cost: Option<u64>,
    /// The intended required parallel threads for the PBKDF operation.
    #[serde(default, rename = "luksPbkdfParallelThreads")]
    pub luks_pbkdf_parallel_threads: Option<u64>,
    /// The sector size to use for the LUKS storage mechanism, in bytes.
    #[serde(default, rename = "luksSectorSize")]
    pub luks_sector_size: Option<u64>,
    /// Controls whether the home area shall be grown automatically.
    #[serde(default, rename = "autoResizeMode")]
    #[derivative(Default(value = "AutoResizeMode::AutoResizeOff"))]
    pub auto_resize_mode: AutoResizeMode,
    /// Configures the free disk space rebalancing weight for the home area.
    #[serde(default, rename = "rebalanceWeight")]
    pub rebalance_weight: Option<u64>,
    /// Configures the authentication rate limiting enforced on the user account.
    #[serde(default, rename = "rateLimitIntervalUSec")]
    pub rate_limit_interval_usec: Option<u64>,
    /// Closely related to rateLimitIntervalUSec, that puts a limit on authentication attempts
    /// within the configured time interval.
    #[serde(default, rename = "rateLimitIntervalBurst")]
    pub rate_limit_interval_burst: Option<u64>,
    /// Configures whether to enforce the system’s password policy when creating the home directory
    /// for the user or changing the user’s password.
    #[serde(default, rename = "enforcePasswordPolicy")]
    #[derivative(Default(value = "true"))]
    pub enforce_password_policy: bool,
    /// If true the user record is marked as suitable for auto-login.
    #[serde(default, rename = "autoLogin")]
    pub auto_login: bool,
    /// The user’s preferred session type (i.e. x11, wayland, or other values valid for
    /// `$XDG_SESSION_TYPE`).
    #[serde(default, rename = "preferredSessionType")]
    pub preferred_session_type: Option<String>,
    /// The user’s preferred session launcher desktop entry file (i.e. gnome, plasma, ...)
    #[serde(default, rename = "preferredSessionLauncher")]
    pub preferred_session_launcher: Option<String>,
    /// The time in µs the per-user service manager is kept around after the user fully logged out.
    #[serde(default, rename = "stopDelayUSec")]
    pub stop_delay_usec: Option<u64>,
    ///  If true all processes of the user are automatically killed when the user logs out.
    #[serde(default, rename = "killProcesses")]
    pub kill_processes: bool,
    /// Encoding how much time has to pass at least between password changes of the user.
    #[serde(default, rename = "passwordChangeMinUSec")]
    pub password_change_min_usec: Option<u64>,
    /// Encodes how much time has to pass at at most between password changes of the user.
    #[serde(default, rename = "passwordChangeMaxUSec")]
    pub password_change_max_usec: Option<u64>,
    /// How much time to warn the user before their password expires, in µs.
    #[serde(default, rename = "passwordChangeWarnUSec")]
    pub password_change_warn_usec: Option<u64>,
    /// Encodes how much time has to pass after the password expired that the account is
    /// deactivated, in µs.
    #[serde(default, rename = "passwordChangeInactiveUSec")]
    pub password_change_inactive_usec: Option<u64>,
    /// If true the user has to change their password on next login.
    #[serde(default, rename = "passwordChangeNow")]
    pub password_change_now: bool,
    /// An array of strings, each with an RFC 7512 compliant PKCS#11 URI referring to security
    /// token (or smart card) of some form, that shall be associated with the user and may be used
    /// for authentication.
    #[serde(default, rename = "pkcs11TokenUri")]
    pub pkcs11_token_uri: Option<Vec<String>>,
    /// An array of strings,each with a Base64-encoded FIDO2 credential ID that shall be used for
    /// authentication with FIDO2 devices that implement the hmac-secret extension.
    #[serde(default, rename = "Fido2HmacCredential")]
    pub fido2_hmac_credential: Option<Vec<String>>,
    /// An array of strings, each corresponding to a field name that can appear in the regular or
    /// perMachine sections.
    #[serde(default, rename = "selfModifiableFields")]
    pub self_modifiable_fields: Option<Vec<String>>,
    /// Similar to selfModifiableFields, but it lists blobs that the user is allowed to edit.
    #[serde(default, rename = "selfModifiableBlobs")]
    pub self_modifiable_blobs: Option<Vec<String>>,
    /// Similar to selfModifiableFields, but it lists fields in the privileged section that the
    /// user is allowed to edit.
    #[serde(default, rename = "selfModifiablePriviliged")]
    pub self_modifiable_privileged: Option<Vec<String>>,
}

/// The binding section overrides their equally named counterparts in the regular and perMachine
/// sections.
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct UserBinding {
    /// The absolute path to a world-readable copy of the user’s blob directory.
    #[serde(default, rename = "blobDirectory")]
    pub blob_directory: Option<String>,
    /// An absolute file system path to the file, directory or block device to use for storage
    /// backing the home directory.
    #[serde(default, rename = "imagePath")]
    pub image_path: Option<String>,
    /// An absolute file system path to the home directory.
    #[serde(default, rename = "homeDirectory")]
    pub home_directory: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the GPT partition UUID the home directory is
    /// located in.
    #[serde(default, rename = "partitionUuid")]
    pub partition_uuid: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the LUKS volume UUID the home directory is
    /// located in.
    #[serde(default, rename = "luksUuid")]
    pub luks_uuid: Option<String>,
    /// One of ext4, xfs, btrfs (possibly others) to use as file system for the user’s home
    /// directory.
    #[serde(default, rename = "fileSystemType")]
    pub file_system_type: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the GPT partition UUID the home directory is
    /// located in.
    #[serde(default, rename = "fileSystemUuid")]
    pub file_system_uuid: Option<String>,
    /// The numeric UNIX user ID (UID) to use for the user.
    #[serde(default)]
    pub uid: Option<u32>,
    /// The numeric UNIX group ID (GID) to use for the user.
    #[serde(default)]
    pub gid: Option<u32>,
    /// The storage mechanism for the user’s home directory.
    #[derivative(Default(value = "UserHomeStorage::UserClassic"))]
    #[serde(default)]
    pub storage: UserHomeStorage,
    /// The cipher to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksCipher")]
    pub luks_cipher: Option<String>,
    // Selects the cipher mode to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksCipherMode")]
    pub luks_cipher_mode: Option<String>,
    /// The volume key length in bytes to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksVolumeKeySize")]
    pub luks_volume_key_size: Option<u64>,
}

/// Contains information about current runtime resource usage.
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct UserStatus {
    /// The currently used disk space of the home directory in bytes.
    #[serde(default, rename = "diskUsage")]
    pub disk_usage: Option<u64>,
    /// The number of 'free' bytes in the disk space allotment.
    #[serde(default, rename = "diskFree")]
    pub disk_free: Option<u64>,
    /// The intended home directory disk space in bytes to assign to the user.
    #[serde(default, rename = "diskSize")]
    pub disk_size: Option<u64>,
    /// The upper bounds when changing the `diskSize` value, in bytes.
    #[serde(default, rename = "diskCeiling")]
    pub disk_ceiling: Option<u64>,
    /// The lower bounds when changing the `diskSize` value, in bytes.
    #[serde(default, rename = "diskFloor")]
    pub disk_floor: Option<u64>,
    /// The current state of the home directory.
    #[serde(default)]
    pub state: Option<String>,
    /// If true indicates that the user record is signed by a public key for which the private key
    /// is available locally.
    #[serde(default, rename = "signedLocally")]
    pub signed_locally: bool,
    /// This counter is increased by one on every successful authentication attempt.
    #[serde(default, rename = "goodAuthenticationCounter")]
    pub good_authentication_counter: Option<u64>,
    /// This counter is increased by one on every unsuccessfully authentication attempt.
    #[serde(default, rename = "badAuthenticationCounter")]
    pub bad_authentication_counter: Option<u64>,
    /// The time of the last successful authentication attempt in µs since the UNIX epoch (1970).
    #[serde(default, rename = "lastGoodAuthenticationUSec")]
    pub last_good_authentication_usec: Option<u64>,
    /// Similar, but the timestamp of the last unsuccessfully authentication attempt.
    #[serde(default, rename = "lastBadAuthenticationUSec")]
    pub last_bad_authentication_usec: Option<u64>,
    /// The µs timestamp since the UNIX epoch (1970) where the most recent rate limiting interval
    /// has been started.
    #[serde(default, rename = "rateLimitBeginUSec")]
    pub rate_limit_begin_usec: Option<u64>,
    /// Counting the authentication attempts in the current rate limiting interval.
    #[serde(default, rename = "rateLimitCount")]
    pub rate_limit_count: Option<u64>,
    /// If true the manager of this user record determined the home directory being on removable
    /// media.
    #[serde(default)]
    pub removeable: bool,
    /// The service that defines or manages this user record.
    #[serde(default)]
    pub service: Option<String>,
}

/// Contains one or more cryptographic signatures of a reduced version of the user record.
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct UserSignature {
    /// The signature, encoded in Base64
    #[serde(default)]
    pub data: String,
    /// A copy of the public key whose private key was used to make the signature, in PEM format.
    #[serde(default)]
    pub key: String,
}

/// Secret user credentials, such as password or PIN information.
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct UserSecret {
    /// An array of strings, each containing a plain text password.
    #[serde(default)]
    pub passwords: Option<Vec<String>>,
    /// An array of strings, each containing a plain text PIN, suitable for unlocking security
    /// tokens that require that.
    #[serde(default)]
    pub token_pin: Option<Vec<String>>,
    /// If set to true allows the receiver to use the PKCS#11 “protected authentication path” for
    /// authenticating the user.
    #[serde(default)]
    pub pkcs11_protected_authentication_path_permitted: bool,
    /// If set to true allows the receiver to use the FIDO2 “user presence” flag.
    #[serde(default)]
    pub fido2_user_precense_permitted: bool,
    /// If set to true allows the receiver to use the FIDO2 “user verification” flag.
    #[serde(default)]
    pub fido2_user_verification_permitted: bool,
}

/// User Records
///
/// For a detailed explanation please visit <https://systemd.io/USER_RECORDS>
#[allow(clippy::struct_excessive_bools)]
#[derive(Deserialize, Derivative)]
#[derivative(Debug, Default)]
pub struct UserRecord {
    //
    // Regular section
    //
    /// The UNIX user name for this record.
    #[serde(rename = "userName")]
    pub user_name: String,
    /// The 'realm' a user is defined in.
    #[serde(default)]
    pub realm: Option<String>,
    /// The absolute path to a world-readable copy of the user’s blob directory.
    #[serde(default, rename = "blobDirectory")]
    pub blob_directory: Option<String>,
    /// An object, which maps valid blob directory filenames to SHA256 hashes formatted as hex
    /// strings.
    #[serde(
        default,
        rename = "blobManifest",
        deserialize_with = "decode_string_hashmap"
    )]
    pub blob_manifest: Option<HashMap<String, String>>,
    /// The real name of the user.
    #[serde(default, rename = "realName")]
    pub real_name: Option<String>,
    /// The email address of the user.
    #[serde(default, rename = "emailAddress")]
    pub email_address: Option<String>,
    /// The name of an icon picked by the user.
    #[serde(default, rename = "iconName")]
    pub icon_name: Option<String>,
    /// A free-form location string describing the location of the user.
    #[serde(default)]
    pub location: Option<String>,
    /// Clarifies the disposition of the user, i.e. the context it is defined in.
    #[derivative(Default(value = "UserDisposition::UserRegular"))]
    #[serde(default)]
    pub disposition: UserDisposition,
    /// A timestamp indicating when the user record was last changed (µs since the epoch 1970).
    #[serde(default, rename = "lastChangeUSec")]
    pub last_change_usec: Option<u64>,
    /// A timestamp indicating the point in time when password of the user was last changed (µs
    /// since the epoch 1970).
    #[serde(default, rename = "lastPasswordChangeUSec")]
    pub last_password_change_usec: Option<u64>,
    /// The shell binary to use for terminal logins of this user.
    #[serde(default)]
    pub shell: Option<String>,
    /// The `umask` to set for the user’s login sessions.
    #[serde(default)]
    pub umask: Option<u32>,
    /// An array of strings, each containing an environment variable and its value to set for the
    /// user’s login session.
    #[serde(default, deserialize_with = "decode_string_hashmap")]
    pub environment: Option<HashMap<String, String>>,
    /// The preferred timezone of the user.
    #[serde(default, rename = "timeZone")]
    pub time_zone: Option<String>,
    /// The preferred language/locale for the user.
    #[serde(default, rename = "preferredLanguage")]
    pub preferred_language: Option<String>,
    /// The preferred languages/locales that should be used in the event that translations for the
    /// `preferredLanguage` are missing
    #[serde(default, rename = "additionalLanguages")]
    pub additional_languages: Option<Vec<String>>,
    /// An integer value in the range -20…19
    #[serde(default, rename = "niceLevel")]
    pub nice_level: Option<i32>,
    // An object, where each key refers to a Linux resource limit (such as RLIMIT_NOFILE and
    // similar).
    // #[serde(default, rename = "resourceLimits")]
    // TODO pub resource_limits: ???,
    /// If true, the user account is locked, the user may not log in.
    #[derivative(Default(value = "false"))]
    #[serde(default)]
    pub locked: bool,
    /// A timestamp when record should be considered invalid for the purpose of logging in (µs
    /// since the UNIX epoch 1970).
    #[serde(default, rename = "notBeforeUSec")]
    pub not_before_usec: Option<u64>,
    /// The point in time after which logins shall not be permitted anymore (µs since the
    /// UNIX epoch 1970).
    #[serde(default, rename = "notAfterUSec")]
    pub not_after_usec: Option<u64>,
    /// The storage mechanism for the user’s home directory.
    #[derivative(Default(value = "UserHomeStorage::UserClassic"))]
    #[serde(default)]
    pub storage: UserHomeStorage,
    /// The intended home directory disk space in bytes to assign to the user.
    #[serde(default, rename = "diskSize")]
    pub disk_size: Option<u64>,
    /// Similar to diskSize but takes a relative value, but specifies a fraction of the available
    /// disk space on the selected storage medium to assign to the user.
    #[serde(default, rename = "diskSizeRelative")]
    pub disk_size_relative: Option<u64>,
    /// The absolute path to the skeleton directory to populate a new home directory from.
    #[serde(default, rename = "skeletonDirectory")]
    pub skeleton_directory: Option<String>,
    /// The UNIX access mask in the range 0…511 for the home directory when it is first created.
    #[serde(default, rename = "accessMask")]
    pub access_mask: Option<u32>,
    /// The maximum number of tasks the user may start in parallel during system runtime.
    #[serde(default, rename = "taskMax")]
    pub tasks_max: Option<u64>,
    /// Upper memory limits for all processes of the user.
    #[serde(default, rename = "memoryHigh")]
    pub memory_high: Option<u64>,
    /// Upper memory limits for all processes of the user.
    #[serde(default, rename = "memoryMax")]
    pub memory_max: Option<u64>,
    /// CPU scheduling weights for the user’s processes (range 1…10000 ,defaults to
    /// 100).
    #[serde(default, rename = "cpuWeight")]
    pub cpu_weight: Option<u32>,
    /// IO scheduling weights for the user’s processes (range 1…10000 ,defaults to 100).
    #[serde(default, rename = "ioWeight")]
    pub io_weight: Option<u32>,
    /// Control nosuid mount flag of the user’s home directories.
    #[serde(default, rename = "mountNoSuid")]
    pub mount_no_suid: bool,
    /// Control nodev mount flag of the user’s home directories.
    #[serde(default, rename = "mountNoDev")]
    pub mount_no_dev: bool,
    /// Control noexec mount flag of the user’s home directories.
    #[serde(default, rename = "mountNoExec")]
    pub mount_no_exec: bool,
    /// The Windows File Sharing domain (CIFS) to use.
    #[serde(default, rename = "cifsDomain")]
    pub cifs_domain: Option<String>,
    /// The Windows File Sharing user name (CIFS) to associate this user record with.
    #[serde(default, rename = "cifsName")]
    pub cifs_user_name: Option<String>,
    /// The Windows File Share service (CIFS) to mount as home directory of the user on login.
    #[serde(default, rename = "cifsService")]
    pub cifs_service: Option<String>,
    /// Additional mount options to pass to mount.cifs when mounting the home directory CIFS share.
    #[serde(default, rename = "cifsExtraMountOptions")]
    pub cifs_extra_mount_options: Option<String>,
    /// An absolute file system path to the file, directory or block device to use for storage
    /// backing the home directory.
    #[serde(default, rename = "imagePath")]
    pub image_path: Option<String>,
    /// An absolute file system path to the home directory.
    #[serde(default, rename = "homeDirectory")]
    pub home_directory: Option<String>,
    /// The numeric UNIX user ID (UID) to use for the user.
    #[serde(default)]
    pub uid: Option<u32>,
    /// The numeric UNIX group ID (GID) to use for the user.
    #[serde(default)]
    pub gid: Option<u32>,
    /// An array of strings, each indicating a UNIX group this user shall be a member of.
    #[serde(default, rename = "memberOf")]
    pub member_of: Option<Vec<String>>,
    /// One of ext4, xfs, btrfs (possibly others) to use as file system for the user’s home
    /// directory.
    #[serde(default, rename = "fileSystemType")]
    pub file_system_type: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the GPT partition UUID the home directory is
    /// located in.
    #[serde(default, rename = "partitionUuid")]
    pub partition_uuid: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the LUKS volume UUID the home directory is
    /// located in.
    #[serde(default, rename = "luksUuid")]
    pub luks_uuid: Option<String>,
    /// A lower-case, text-formatted UUID, referencing the file system UUID the home directory is
    /// located in.
    #[serde(default, rename = "fileSystemUuid")]
    pub file_system_uuid: Option<String>,
    ///  If true and luks storage is used, controls whether the loopback block devices, LUKS and
    ///  the file system on top shall be used in discard mode.
    #[serde(default, rename = "luksDiscard")]
    pub luks_discard: bool,
    /// Similar to luksDiscard, it controls whether to trim/allocate the file system/backing file
    /// when deactivating the home directory.
    #[serde(default, rename = "luksOfflineDiscard")]
    pub luks_offline_discard: bool,
    /// Additional mount options to append to the default mount options for the file system in the
    /// LUKS volume.
    #[serde(default, rename = "luksExtraMountOptions")]
    pub luks_extra_mount_options: Option<String>,
    /// The cipher to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksCipher")]
    pub luks_cipher: Option<String>,
    // Selects the cipher mode to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksCipherMode")]
    pub luks_cipher_mode: Option<String>,
    /// The volume key length in bytes to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksVolumeKeySize")]
    pub luks_volume_key_size: Option<u64>,
    /// Selects the hash algorithm to use for the PBKDF operation.
    #[serde(default, rename = "luksPbkdfHashAlgorithm")]
    pub luks_pbkdf_hash_algorithm: Option<String>,
    /// The PBKDF type to use for the LUKS storage mechanism.
    #[serde(default, rename = "luksPbkdfType")]
    pub luks_pbkdf_type: Option<String>,
    /// The intended number of iterations for the PBKDF operation.
    #[serde(default, rename = "luksPbkdfForceIterations")]
    pub luks_pbkdf_force_iterations: Option<u64>,
    /// The intended time cost for the PBKDF operation, in µs.
    #[serde(default, rename = "luksPbkdfTimeCostUSec")]
    pub luks_pbkdf_time_cost_usec: Option<u64>,
    /// The intended memory cost for the PBKDF operation, in bytes.
    #[serde(default, rename = "luksPbkdfMemoryCost")]
    pub luks_pbkdf_memory_cost: Option<u64>,
    /// The intended required parallel threads for the PBKDF operation.
    #[serde(default, rename = "luksPbkdfParallelThreads")]
    pub luks_pbkdf_parallel_threads: Option<u64>,
    /// The sector size to use for the LUKS storage mechanism, in bytes.
    #[serde(default, rename = "luksSectorSize")]
    pub luks_sector_size: Option<u64>,
    /// Controls whether the home area shall be grown automatically.
    #[serde(default, rename = "autoResizeMode")]
    #[derivative(Default(value = "AutoResizeMode::AutoResizeOff"))]
    pub auto_resize_mode: AutoResizeMode,
    /// Configures the free disk space rebalancing weight for the home area.
    #[serde(default, rename = "rebalanceWeight")]
    pub rebalance_weight: Option<u64>,
    /// The service that defines or manages this user record.
    #[serde(default)]
    pub service: Option<String>,
    /// Configures the authentication rate limiting enforced on the user account.
    #[serde(default, rename = "rateLimitIntervalUSec")]
    pub rate_limit_interval_usec: Option<u64>,
    /// Closely related to rateLimitIntervalUSec, that puts a limit on authentication attempts
    /// within the configured time interval.
    #[serde(default, rename = "rateLimitIntervalBurst")]
    pub rate_limit_interval_burst: Option<u64>,
    /// Configures whether to enforce the system’s password policy when creating the home directory
    /// for the user or changing the user’s password.
    #[serde(default, rename = "enforcePasswordPolicy")]
    #[derivative(Default(value = "true"))]
    pub enforce_password_policy: bool,
    /// If true the user record is marked as suitable for auto-login.
    #[serde(default, rename = "autoLogin")]
    pub auto_login: bool,
    /// The user’s preferred session type (i.e. x11, wayland, or other values valid for
    /// `$XDG_SESSION_TYPE`).
    #[serde(default, rename = "preferredSessionType")]
    pub preferred_session_type: Option<String>,
    /// The user’s preferred session launcher desktop entry file (i.e. gnome, plasma, ...)
    #[serde(default, rename = "preferredSessionLauncher")]
    pub preferred_session_launcher: Option<String>,
    /// The time in µs the per-user service manager is kept around after the user fully logged out.
    #[serde(default, rename = "stopDelayUSec")]
    pub stop_delay_usec: Option<u64>,
    ///  If true all processes of the user are automatically killed when the user logs out.
    #[serde(default, rename = "killProcesses")]
    pub kill_processes: bool,
    /// Encoding how much time has to pass at least between password changes of the user.
    #[serde(default, rename = "passwordChangeMinUSec")]
    pub password_change_min_usec: Option<u64>,
    /// Encodes how much time has to pass at at most between password changes of the user.
    #[serde(default, rename = "passwordChangeMaxUSec")]
    pub password_change_max_usec: Option<u64>,
    /// How much time to warn the user before their password expires, in µs.
    #[serde(default, rename = "passwordChangeWarnUSec")]
    pub password_change_warn_usec: Option<u64>,
    /// Encodes how much time has to pass after the password expired that the account is
    /// deactivated, in µs.
    #[serde(default, rename = "passwordChangeInactiveUSec")]
    pub password_change_inactive_usec: Option<u64>,
    /// If true the user has to change their password on next login.
    #[serde(default, rename = "passwordChangeNow")]
    pub password_change_now: bool,
    /// An array of strings, each with an RFC 7512 compliant PKCS#11 URI referring to security
    /// token (or smart card) of some form, that shall be associated with the user and may be used
    /// for authentication.
    #[serde(default, rename = "pkcs11TokenUri")]
    pub pkcs11_token_uri: Option<Vec<String>>,
    /// An array of strings, each with a Base64-encoded FIDO2 credential ID that shall be used for
    /// authentication with FIDO2 devices that implement the hmac-secret extension.
    #[serde(default, rename = "Fido2HmacCredential")]
    pub fido2_hmac_credential: Option<Vec<String>>,
    /// An array of strings, each indicating the type of one recovery key.
    #[serde(default, rename = "recoveryKeyType")]
    pub recovery_key_type: Option<Vec<String>>,
    /// An array of strings, each corresponding to a field name that can appear in the regular or
    /// perMachine sections.
    #[serde(default, rename = "selfModifiableFields")]
    pub self_modifiable_fields: Option<Vec<String>>,
    /// Similar to selfModifiableFields, but it lists blobs that the user is allowed to edit.
    #[serde(default, rename = "selfModifiableBlobs")]
    pub self_modifiable_blobs: Option<Vec<String>>,
    /// Similar to selfModifiableFields, but it lists fields in the privileged section that the
    /// user is allowed to edit.
    #[serde(default, rename = "selfModifiablePriviliged")]
    pub self_modifiable_privileged: Option<Vec<String>>,

    //
    // Privileged section
    //
    #[serde(default)]
    pub privileged: Option<UserPrivileged>,

    //
    // PerMachine section
    //
    pub per_machine: Option<Vec<UserPerMachine>>,

    //
    // Binding section
    //
    #[serde(default)]
    pub binding: Option<HashMap<String, UserBinding>>,

    //
    // Status section
    //
    #[serde(default)]
    pub status: Option<HashMap<String, UserStatus>>,

    //
    // Signature section
    //
    #[serde(default)]
    pub signature: Option<Vec<UserSignature>>,

    //
    // Secret section
    //
    pub secret: Option<UserSecret>,
}

impl UserRecord {
    fn validate(&self) -> crate::Result<bool> {
        if self.user_name.is_empty() {
            error!("No user name specified!");
            return Err(crate::Error::InvalidRecord);
        }

        // TODO root & nodbody are intrinsic

        Ok(true)
    }
}

fn decode_string_hashmap<'de, D>(d: D) -> Result<Option<HashMap<String, String>>, D::Error>
where
    D: Deserializer<'de>,
{
    let j: Value = Deserialize::deserialize(d)?;

    let map: serde_json::Map<String, Value> = match j {
        Value::Object(map) => map,
        _ => return Err(serde::de::Error::custom("Expected serde_json::Map")),
    };

    let mut m: HashMap<String, String> = HashMap::new();
    for (k, v) in map {
        match v {
            Value::String(s) => {
                m.insert(k, s);
            },
            _ => {
                return Err(serde::de::Error::custom(
                    "Expected serde_json::Value::String",
                ));
            },
        }
    }

    Ok(Some(m))
}

// Trait to convert from json Value to UserRecord
impl TryFrom<serde_json::Value> for UserRecord {
    type Error = crate::Error;

    fn try_from(value: Value) -> Result<Self, Self::Error> {
        let user_record: UserRecord = match serde_json::from_value(value) {
            Ok(u) => u,
            Err(e) => {
                error!("Failed to deserialize serde_json::Value: {e:?}");
                return Err(crate::Error::InvalidRecord);
            },
        };

        // Run sanity checks
        user_record.validate()?;

        Ok(user_record)
    }
}

impl core::fmt::Display for UserRecord {
    fn fmt(
        &self,
        fmt: &mut core::fmt::Formatter<'_>,
    ) -> core::result::Result<(), core::fmt::Error> {
        writeln!(fmt, "User Name:   {}", &self.user_name)?;
        writeln!(fmt, "Disposition: {:?}", &self.disposition)?;
        writeln!(
            fmt,
            "Login OK?:   {}",
            if self.locked { "no" } else { "yes" }
        )?;
        if let Some(uid) = &self.uid {
            writeln!(fmt, "UID:         {uid}")?;
        }
        if let Some(gid) = &self.gid {
            writeln!(fmt, "GID:         {gid}")?;
        }
        if let Some(member_of) = &self.member_of {
            write!(fmt, "Groups:      ")?;
            let mut first = true;
            for member in member_of {
                if first {
                    writeln!(fmt, "{member}")?;
                    first = false;
                } else {
                    writeln!(fmt, "             {member}")?;
                }
            }
        }
        writeln!(fmt, "Storage:     {:?}", &self.storage)?;
        if let Some(shell) = &self.shell {
            writeln!(fmt, "Shell:       {shell}")?;
        }
        if let Some(home) = &self.home_directory {
            writeln!(fmt, "Directory:   {home}")?;
        }

        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use serde_json::json;
    use tracing::Level;
    use tracing_subscriber::FmtSubscriber;

    fn setup() {
        let lvl = Level::WARN;
        let subscriber = FmtSubscriber::builder().with_max_level(lvl).finish();
        tracing::subscriber::set_global_default(subscriber).unwrap();
    }

    #[test]
    fn user_record_validate_test() {
        let u = UserRecord {
            user_name: String::from("hanswurst"),
            ..Default::default()
        };

        assert!(u.validate().is_ok());
    }

    #[test]
    fn user_record_validate_fail_test() {
        let u = UserRecord {
            ..Default::default()
        };

        assert!(u.validate().is_err());
    }

    #[test]
    #[allow(clippy::too_many_lines)]
    fn user_record_tryfrom_homed_json_test() {
        setup();

        let j = json!({
                "autoLogin" : true,
                "binding" : {
                        "15e19cf24e004b949ddaac60c74aa165" : {
                                "fileSystemType" : "ext4",
                                "fileSystemUuid" : "758e88c8-5851-4a2a-b88f-e7474279c111",
                                "gid" : 60232u32,
                                "homeDirectory" : "/home/grobie",
                                "blobDirectory" : "/var/cache/systemd/homed/grobie/",
                                "imagePath" : "/home/grobie.home",
                                "luksCipher" : "aes",
                                "luksCipherMode" : "xts-plain64",
                                "luksUuid" : "e63581ba-79fb-4226-b9de-1888393f7573",
                                "luksVolumeKeySize" : 32,
                                "partitionUuid" : "41f9ce04-c827-4b74-a981-c669f93eb4dc",
                                "storage" : "luks",
                                "uid" : 60232u32
                        }
                },
                "blobManifest" : {
                        "avatar" : "c0636851d25a62d817ff7da4e081d1e646e42c74d0ecb53425f75fcf1ba43b52",
                        "login-background" : "da7ad0222a6edbc6cd095149c72d38d92fd3114f606e4b57469857ef47fade18"
                },
                "disposition" : "regular",
                "enforcePasswordPolicy" : false,
                "lastChangeUSec" : 1_565_950_024_279_735_u64,
                "memberOf" : [
                        "wheel"
                ],
                "privileged" : {
                        "hashedPassword" : [
                                "$6$WHBKvAFFT9jKPA4k$OPY4D4TczKN/jOnJzy54DDuOOagCcvxxybrwMbe1SVdm.Bbr.zOmBdATp.QrwZmvqyr8/SafbbQu.QZ2rRvDs/"
                        ]
                },
                "signature" : [
                        {
                                "data" : "LU/HeVrPZSzi3MJ0PVHwD5m/xf51XDYCrSpbDRNBdtF4fDVhrN0t2I2OqH/1yXiBidXlV0ptMuQVq8KVICdEDw==",
                                "key" : "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEA/QT6kQWOAMhDJf56jBmszEQQpJHqDsGDMZOdiptBgRk=\n-----END PUBLIC KEY-----\n"
                        }
                ],
                "userName" : "grobie",
                "status" : {
                        "15e19cf24e004b949ddaac60c74aa165" : {
                                "goodAuthenticationCounter" : 16,
                                "lastGoodAuthenticationUSec" : 1_566_309_343_044_322_u64,
                                "rateLimitBeginUSec" : 1_566_309_342_340_723_u64,
                                "rateLimitCount" : 1,
                                "state" : "inactive",
                                "service" : "io.systemd.Home",
                                "diskSize" : 161_118_667_776_u64,
                                "diskCeiling" : 190_371_729_408_u64,
                                "diskFloor" : 5_242_880_u64,
                                "signedLocally" : true
                        }
                }
        });

        let u: UserRecord = j.try_into().unwrap();

        // regular section
        assert_eq!(u.user_name, "grobie");

        assert!(u.auto_login);
        assert!(u.blob_manifest.is_some());
        let bm = &u.blob_manifest.unwrap();
        assert_eq!(
            bm["avatar"],
            "c0636851d25a62d817ff7da4e081d1e646e42c74d0ecb53425f75fcf1ba43b52"
        );
        assert_eq!(
            bm["login-background"],
            "da7ad0222a6edbc6cd095149c72d38d92fd3114f606e4b57469857ef47fade18"
        );
        assert!(matches!(u.disposition, UserDisposition::UserRegular));
        assert!(!u.enforce_password_policy);
        assert_eq!(u.last_change_usec, Some(1_565_950_024_279_735_u64));

        assert!(u.member_of.is_some());
        let mem = &u.member_of.unwrap();
        assert_eq!(mem.len(), 1);
        assert_eq!(mem[0], "wheel");

        // privileged section
        assert!(u.privileged.is_some());
        let p = u.privileged.unwrap();
        assert!(p.hashed_passwords.is_some());
        let hp = &p.hashed_passwords.unwrap();
        assert_eq!(
            hp[0],
            "$6$WHBKvAFFT9jKPA4k$OPY4D4TczKN/jOnJzy54DDuOOagCcvxxybrwMbe1SVdm.Bbr.zOmBdATp.QrwZmvqyr8/SafbbQu.QZ2rRvDs/"
        );

        // binding section
        assert!(u.binding.is_some());
        let b = u.binding.unwrap();
        assert!(b.contains_key("15e19cf24e004b949ddaac60c74aa165"));
        let bk = b.get("15e19cf24e004b949ddaac60c74aa165").unwrap();
        assert_eq!(bk.file_system_type, Some("ext4".to_string()));
        assert_eq!(
            bk.file_system_uuid,
            Some("758e88c8-5851-4a2a-b88f-e7474279c111".to_string())
        );
        assert_eq!(bk.gid, Some(60232u32));
        assert_eq!(bk.home_directory, Some("/home/grobie".to_string()));
        assert_eq!(
            bk.blob_directory,
            Some("/var/cache/systemd/homed/grobie/".to_string())
        );
        assert_eq!(bk.image_path, Some("/home/grobie.home".to_string()));
        assert_eq!(bk.luks_cipher, Some("aes".to_string()));
        assert_eq!(bk.luks_cipher_mode, Some("xts-plain64".to_string()));
        assert_eq!(
            bk.luks_uuid,
            Some("e63581ba-79fb-4226-b9de-1888393f7573".to_string())
        );
        assert_eq!(bk.luks_volume_key_size, Some(32));
        assert_eq!(
            bk.partition_uuid,
            Some("41f9ce04-c827-4b74-a981-c669f93eb4dc".to_string())
        );
        assert!(matches!(bk.storage, UserHomeStorage::UserLuks));
        assert_eq!(bk.uid, Some(60232u32));

        // signature section
        assert!(u.signature.is_some());
        let sigv = &u.signature.unwrap();
        assert_eq!(sigv.len(), 1);
        let sig = &sigv[0];
        assert_eq!(
            sig.data,
            "LU/HeVrPZSzi3MJ0PVHwD5m/xf51XDYCrSpbDRNBdtF4fDVhrN0t2I2OqH/1yXiBidXlV0ptMuQVq8KVICdEDw=="
        );
        assert_eq!(
            sig.key,
            "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEA/QT6kQWOAMhDJf56jBmszEQQpJHqDsGDMZOdiptBgRk=\n-----END PUBLIC KEY-----\n"
        );

        // status section
        assert!(u.status.is_some());
        let st = u.status.unwrap();
        assert!(st.contains_key("15e19cf24e004b949ddaac60c74aa165"));
        let stk = st.get("15e19cf24e004b949ddaac60c74aa165").unwrap();
        assert_eq!(stk.good_authentication_counter, Some(16));
        assert_eq!(
            stk.last_good_authentication_usec,
            Some(1_566_309_343_044_322_u64)
        );
        assert_eq!(stk.rate_limit_begin_usec, Some(1_566_309_342_340_723_u64));
        assert_eq!(stk.rate_limit_count, Some(1));
        assert_eq!(stk.disk_size, Some(161_118_667_776_u64));
        assert_eq!(stk.disk_ceiling, Some(190_371_729_408_u64));
        assert_eq!(stk.disk_floor, Some(5_242_880_u64));
        assert!(stk.signed_locally);
    }
}
